@@ -1,5 +1,5 @@
 /*
- * $Id: msqltcl.c,v 1.0 1995/01/05 11:13:05 hs Rel $
+ * $Id: msqltcl.c,v 1.50 1995/06/06 14:25:25 hs Rel $
  *
  * MSQL interface to Tcl
  *
@@ -37,7 +37,7 @@
 
 
 #define MSQL_HANDLES      15	/* Default number of handles available. */
-#define MSQL_BUFF_SIZE	4096	/* Conversion buffer size for various needs. */
+#define MSQL_BUFF_SIZE	1024	/* Conversion buffer size for various needs. */
 #define MSQL_SMALL_SIZE  TCL_RESULT_SIZE /* Smaller buffer size. */
 #define MSQL_NAME_LEN     80    /* Max. host, database name length. */
 
@@ -58,7 +58,16 @@ static char *MsqlHandlePrefix = "msql";
  */
 #define MSQL_HPREFIX_LEN 4
 
-static char *MsqlStatusArr = "msqlstatus";  /* Array for status info. */
+/* Array for status info, and its elements. */
+static char *MsqlStatusArr = "msqlstatus";
+#define MSQL_STATUS_CODE "code"
+#define MSQL_STATUS_CMD  "command"
+#define MSQL_STATUS_MSG  "message"
+#define MSQL_STATUS_NULLV  "nullvalue"
+
+/* C variable corresponding to msqlstatus(nullvalue) */
+static char* MsqlNullvalue = NULL ;
+#define MSQL_NULLV_INIT ""
 
 /* Options to the 'info', 'result', 'col' combo commands. */
      
@@ -77,14 +86,16 @@ static char* MsqlDbOpt[] =
 
 static char* MsqlResultOpt[] =
 {
-  "rows", "rows?", "cols", "cols?"
+  "rows", "rows?", "cols", "cols?", "current", "current?"
 };
 #define MSQL_RESROWS_OPT 0
 #define MSQL_RESROWSQ_OPT 1
 #define MSQL_RESCOLS_OPT 2
 #define MSQL_RESCOLSQ_OPT 3
+#define MSQL_RESCUR_OPT 4
+#define MSQL_RESCURQ_OPT 5
 
-#define MSQL_RES_OPT_MAX 3
+#define MSQL_RES_OPT_MAX 5
 
 /* Column info definitions. */
 
@@ -124,6 +135,7 @@ extern Tcl_CmdProc  Msqltcl_Connect;
 extern Tcl_CmdProc  Msqltcl_Use;
 extern Tcl_CmdProc  Msqltcl_Sel;
 extern Tcl_CmdProc  Msqltcl_Next;
+extern Tcl_CmdProc  Msqltcl_Seek;
 extern Tcl_CmdProc  Msqltcl_Map;
 extern Tcl_CmdProc  Msqltcl_Exec;
 extern Tcl_CmdProc  Msqltcl_Close;
@@ -132,64 +144,125 @@ extern Tcl_CmdProc  Msqltcl_Result;
 extern Tcl_CmdProc  Msqltcl_Col;
 extern Tcl_CmdProc  Msqltcl_State;
 
+  
+/* CONFLICT HANDLING
+ *
+ * Every command begins by calling 'msql_prologue'.
+ * This function resets msqlstatus(code) to zero; the other array elements
+ * retain their previous values.
+ * The function also saves argc/argv in global variables.
+ * After this the command processing proper begins.
+ *
+ * If there is a conflict, the message is taken from one of the following
+ * sources,
+ * -- this code (msql_prim_confl),
+ * -- the database server (msql_server_confl),
+ * -- a POSIX (system call) diagnostic.
+ * A complete message is put together from the above plus the name of the
+ * command where the conflict was detected.
+ * The complete message is returned as the Tcl result and is also stored in
+ * msqlstatus(message).
+ * msqlstatus(code) is set to "-1", except for POSIX conflicts where 'errno'
+ * is used.
+ * In addition, the whole command where the conflict was detected is put
+ * together from the saved argc/argv and is copied into msqlstatus(command).
+ */
+
+static Tcl_Interp* saved_interp;
+static int saved_argc;
+static char** saved_argv;
+
 
 /*
  *----------------------------------------------------------------------
- * msql_tcl_conflict
- * Sets the status array after an mSQL conflict.
- *
+ * msql_reassemble
+ * Reassembles the current command from the saved argv; copies it into
+ * msqlstatus(command).
  */
 
 static void
-msql_tcl_conflict (interp)
-     Tcl_Interp *interp ;
+msql_reassemble ()
 {
-    Tcl_SetVar2(interp, MsqlStatusArr, "code", "-1", TCL_GLOBAL_ONLY);
-    Tcl_SetVar2(interp, MsqlStatusArr, "message",
-		(msqlErrMsg == NULL) ? "" : msqlErrMsg, TCL_GLOBAL_ONLY);
+  unsigned int flags = TCL_GLOBAL_ONLY | TCL_LIST_ELEMENT;
+  int idx ;
+
+  for (idx = 0; idx < saved_argc; ++idx)
+    {
+      Tcl_SetVar2 (saved_interp, MsqlStatusArr, MSQL_STATUS_CMD,
+		   saved_argv[idx], flags) ;
+      flags |= TCL_APPEND_VALUE ;
+    }
 }
 
 
 /*
  *----------------------------------------------------------------------
- * msql_prim_conflict
- * Sets the status array after a primitive conflict.
- * If 'mess' is NULL the current result is used.
- * RETURNS 'TCL_ERROR' for convenience.
+ * msql_posix_confl
+ * Conflict handling after a failed system call.
+ */
+
+static int
+msql_posix_confl (syscall)
+     char* syscall;
+{
+  char buf[MSQL_SMALL_SIZE];
+  char* strerror () ;
+  char* msg = strerror (errno) ;
+
+  sprintf(buf, "%d", errno) ;
+  Tcl_SetVar2 (saved_interp, MsqlStatusArr, MSQL_STATUS_CODE, buf,
+	       TCL_GLOBAL_ONLY);
+  Tcl_SetResult (saved_interp, "", TCL_STATIC) ;
+  Tcl_AppendResult (saved_interp, saved_argv[0], "/POSIX: (", syscall,
+		    ") ", (msg == NULL) ? "" : msg, (char*)NULL);
+  Tcl_SetVar2 (saved_interp, MsqlStatusArr, MSQL_STATUS_MSG,
+	       saved_interp->result, TCL_GLOBAL_ONLY);
+  msql_reassemble () ;
+  return TCL_ERROR ;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ * msql_prim_confl
+ * Conflict handling after a primitive conflict.
  *
  */
 
 static int
-msql_prim_conflict (interp, mess)
-     Tcl_Interp *interp ;
-     char* mess ;
+msql_prim_confl (msg)
+     char* msg ;
 {
-    Tcl_SetVar2(interp, MsqlStatusArr, "code", "-1", TCL_GLOBAL_ONLY);
-    Tcl_SetVar2(interp, MsqlStatusArr, "message",
-		(mess == NULL) ? interp->result : mess, TCL_GLOBAL_ONLY);
-    return TCL_ERROR ;
+  Tcl_SetVar2 (saved_interp, MsqlStatusArr, MSQL_STATUS_CODE, "-1",
+	       TCL_GLOBAL_ONLY);
+  Tcl_SetResult (saved_interp, "", TCL_STATIC) ;
+  Tcl_AppendResult (saved_interp, saved_argv[0], ": ", msg, (char*)NULL) ;
+  Tcl_SetVar2 (saved_interp, MsqlStatusArr, MSQL_STATUS_MSG,
+	       saved_interp->result, TCL_GLOBAL_ONLY);
+  msql_reassemble () ;
+  return TCL_ERROR ;
 }
 
-  
+
 /*
  *----------------------------------------------------------------------
- * msql_tcl_perror
- * Sets the status array after a Unix system call conflict.
+ * msql_server_confl
+ * Conflict handling after an mSQL conflict.
  *
  */
 
-static void
-msql_tcl_perror (interp)
-     Tcl_Interp *interp ;
+static int
+msql_server_confl ()
 {
-    char buf[MSQL_BUFF_SIZE];
-    char* strerror () ;
-    char* msg = strerror (errno) ;
-
-    sprintf(buf, "%d", errno) ;
-    Tcl_SetVar2(interp, MsqlStatusArr, "code", buf, TCL_GLOBAL_ONLY);
-    Tcl_SetVar2(interp, MsqlStatusArr, "message",
-		(msg == NULL) ? "" : msg, TCL_GLOBAL_ONLY);
+  Tcl_SetVar2 (saved_interp, MsqlStatusArr, MSQL_STATUS_CODE, "-1",
+	       TCL_GLOBAL_ONLY);
+  Tcl_SetResult (saved_interp, "", TCL_STATIC) ;
+  Tcl_AppendResult (saved_interp, saved_argv[0], "/db server: ",
+		    (msqlErrMsg == NULL) ? "" : msqlErrMsg, (char*)NULL) ;
+  Tcl_SetVar2 (saved_interp, MsqlStatusArr, MSQL_STATUS_MSG,
+	       saved_interp->result, TCL_GLOBAL_ONLY);
+  msql_reassemble () ;
+  return TCL_ERROR ;
 }
 
 
@@ -199,16 +272,15 @@ msql_tcl_perror (interp)
  * RETURN: MsqlHandle index number or -1 on error.
  */
 static int
-get_handle_plain (interp, handle) 
-     Tcl_Interp *interp ;
+get_handle_plain (handle) 
      char *handle;
 {
   int hi ;
   char* hp = handle ;
-
+  
   if (HSYNTAX(hp,hi) < 0)
     {
-      msql_prim_conflict (interp, "Weird handle") ;
+      msql_prim_confl ("weird handle") ; /*  */
       return -1 ;
     }
   else
@@ -222,8 +294,7 @@ get_handle_plain (interp, handle)
  * RETURN: MsqlHandle index number or -1 on error.
  */
 static int
-get_handle_conn (interp, handle) 
-     Tcl_Interp *interp ;
+get_handle_conn (handle) 
      char *handle;
 {
   int hi ;
@@ -231,13 +302,13 @@ get_handle_conn (interp, handle)
 
   if (HSYNTAX(hp,hi) < 0)
     {
-      msql_prim_conflict (interp, "Weird handle") ;
+      msql_prim_confl ("weird handle") ;
       return -1 ;
     }
 
   if (MsqlHandle[hi].connection < 0)
     {
-      msql_prim_conflict (interp, "Handle not connected") ;
+      msql_prim_confl ("handle not connected") ;
       return -1 ;
     }
   else
@@ -252,8 +323,7 @@ get_handle_conn (interp, handle)
  * RETURN: MsqlHandle index number or -1 on error.
  */
 static int
-get_handle_db (interp, handle) 
-     Tcl_Interp *interp ;
+get_handle_db (handle) 
      char *handle;
 {
   int hi ;
@@ -261,19 +331,19 @@ get_handle_db (interp, handle)
 
   if (HSYNTAX(hp,hi) < 0)
     {
-      msql_prim_conflict (interp, "Weird handle") ;
+      msql_prim_confl ("weird handle") ;
       return -1 ;
     }
 
   if (MsqlHandle[hi].connection < 0)
     {
-      msql_prim_conflict (interp, "Handle not connected") ;
+      msql_prim_confl ("handle not connected") ;
       return -1 ;
     }
 
   if (MsqlHandle[hi].database[0] == '\0')
     {
-      msql_prim_conflict (interp, "No current database") ;
+      msql_prim_confl ("no current database") ;
       return -1 ;
     }
   else
@@ -287,9 +357,9 @@ get_handle_db (interp, handle)
  * there is a current database and that there is a pending result.
  * RETURN: MsqlHandle index number or -1 on error.
  */
+
 static int
-get_handle_res (interp, handle) 
-     Tcl_Interp *interp ;
+get_handle_res (handle) 
      char *handle;
 {
   int hi ;
@@ -297,25 +367,25 @@ get_handle_res (interp, handle)
 
   if (HSYNTAX(hp,hi) < 0)
     {
-      msql_prim_conflict (interp, "Weird handle") ;
+      msql_prim_confl ("weird handle") ;
       return -1 ;
     }
 
   if (MsqlHandle[hi].connection < 0)
     {
-      msql_prim_conflict (interp, "Handle not connected") ;
+      msql_prim_confl ("handle not connected") ;
       return -1 ;
     }
 
   if (MsqlHandle[hi].database[0] == '\0')
     {
-      msql_prim_conflict (interp, "No current database") ;
+      msql_prim_confl ("no current database") ;
       return -1 ;
     }
 
   if (MsqlHandle[hi].result == NULL)
     {
-      msql_prim_conflict (interp, "No result pending") ;
+      msql_prim_confl ("no result pending") ; /*  */
       return -1 ;
     }
   else
@@ -356,8 +426,9 @@ static void
 clear_msg(interp)
     Tcl_Interp *interp;
 {
-    Tcl_SetVar2(interp, MsqlStatusArr, "code", "0", TCL_GLOBAL_ONLY);
-    Tcl_SetVar2(interp, MsqlStatusArr, "message", "", TCL_GLOBAL_ONLY);
+    Tcl_SetVar2(interp, MsqlStatusArr, MSQL_STATUS_CODE, "0", TCL_GLOBAL_ONLY);
+    Tcl_SetVar2(interp, MsqlStatusArr, MSQL_STATUS_CMD, "", TCL_GLOBAL_ONLY);
+    Tcl_SetVar2(interp, MsqlStatusArr, MSQL_STATUS_MSG, "", TCL_GLOBAL_ONLY);
 }
 
 
@@ -365,46 +436,55 @@ clear_msg(interp)
  *----------------------------------------------------------------------
  * msql_prologue
  *
- * Does most of standard command prologue, assumes handle is argv[1].
- * RETURNS: Handle index  or -1 on failure.
- * 
+ * Does most of standard command prologue; required for all commands
+ * having conflict handling.
+ * 'req_args' must be the required number of arguments for the command,
+ * including the command word.
+ * 'usage_msg' must be a usage message, leaving out the command name.
+ * Checks the handle assumed to be present in argv[1] if 'check' is not NULL.
+ * RETURNS: Handle index or -1 on failure.
+ * Returns zero if 'check' is NULL.
+ * SIDE EFFECT: Sets the Tcl result on failure.
  */
 
 static int
-msql_prologue (interp, argc, argv, num_args, err_msg, check)
+msql_prologue (interp, argc, argv, req_args, check, usage_msg)
      Tcl_Interp *interp;
      int         argc;
      char      **argv;
-     int         num_args;
-     char       *err_msg;
+     int         req_args;
+     char       *usage_msg;
      int (*check) () ; /* Pointer to function for checking the handle. */
 {
-    int         hand;
+  char buf[MSQL_BUFF_SIZE];
+  int hand = 0;
+  int need;
 
-    /* Check number of minimum args. */
+  /* Reset msqlstatus(code). */
+  Tcl_SetVar2 (interp, MsqlStatusArr, MSQL_STATUS_CODE, "0",
+	       TCL_GLOBAL_ONLY);
 
-    if (argc < num_args)
-      {
-	Tcl_AppendResult (interp, "usage: ", argv[0], err_msg, (char*)NULL);
-	(void)msql_prim_conflict (interp, (char*)NULL) ;
-	return -1 ;
-      }
+  /* Save command environment. */
+  saved_interp = interp;
+  saved_argc = argc ;
+  saved_argv = argv ;
 
-    /* Check the handle.
-     * The function is assumed to set the status array on conflict.
-     */
+  /* Check number of minimum args. */
+  if ((need = req_args - argc) > 0) 
+    {
+      sprintf (buf, "%d more %s needed: %s %s", need, (need>1)?"args":"arg",
+	       argv[0], usage_msg);
+      (void)msql_prim_confl (buf) ;
+      return -1 ;
+    }
 
-    if ((hand = check (interp, argv[1])) < 0)
-      {
-	Tcl_AppendResult (interp, argv[0], " ", argv[1],
-			  ": invalid handle ", (char*)NULL);
-	return -1 ;
-      }
+  /* Check the handle.
+   * The function is assumed to set the status array on conflict.
+   */
+  if (check != NULL && (hand = check (argv[1])) < 0)
+    return -1 ;
 
-    /* Clear msqlstatus array for new messages & errors. */
-    clear_msg(interp);
-
-    return (hand);
+  return (hand);
 }
 
 
@@ -415,8 +495,8 @@ msql_prologue (interp, argc, argv, num_args, err_msg, check)
  * Given an m_field struct and a string keyword appends a piece of
  * column info (one item) to the Tcl result.
  * ASSUMES 'fld' is non-null.
- * In case of conflict, replaces the result by a message.
  * RETURNS 0 on success, 1 otherwise.
+ * SIDE EFFECT: Sets the result and status on failure.
  */
 
 static int
@@ -430,7 +510,7 @@ msql_colinfo (interp, fld, keyw)
   int idx ;
   char* res ;
   int retcode ;
-
+  
   for (idx = 0;
        idx <= MSQL_COL_K_MAX && strcmp (MsqlColkey[idx], keyw) != 0;
        idx++) ;
@@ -456,7 +536,7 @@ msql_colinfo (interp, fld, keyw)
 	  res = "real" ;
 	  break ;
 	default:
-	  sprintf (buf, "msqlcol: column '%s' has weird datatype", fld->name) ;
+	  sprintf (buf, "column '%s' has weird datatype", fld->name) ;
 	  res = NULL ;
 	}
       break ;
@@ -481,14 +561,13 @@ msql_colinfo (interp, fld, keyw)
       else
 	strcpy (keybuf, keyw) ;
 
-      sprintf (buf, "msqlcol: unknown option: %s", keybuf) ;
+      sprintf (buf, "unknown option: %s", keybuf) ;
       res = NULL ;
     }
 
   if (res == NULL)
     {
-      Tcl_SetResult (interp, buf, TCL_VOLATILE) ;
-      (void)msql_prim_conflict (interp, (char*)NULL) ;
+      (void)msql_prim_confl (buf) ;
       retcode = 1 ;
     }
   else
@@ -540,6 +619,7 @@ Msqltcl_Init (interp)
     Tcl_Interp *interp;
 {
   int i;
+  char nbuf[MSQL_SMALL_SIZE];
 
   /*
    * Initialize mSQL proc structures 
@@ -547,7 +627,8 @@ Msqltcl_Init (interp)
   handle_init () ;
 
   /*
-   * Initialize the new Tcl commands
+   * Initialize the new Tcl commands.
+   * Deleting any command will close all connections.
    */
   Tcl_CreateCommand (interp, "msqlconnect", Msqltcl_Connect, (ClientData)NULL,
 		     Msqltcl_Kill);
@@ -557,6 +638,8 @@ Msqltcl_Init (interp)
 		     Msqltcl_Kill);
   Tcl_CreateCommand (interp, "msqlnext",    Msqltcl_Next,    (ClientData)NULL,
 		     Msqltcl_Kill);
+  Tcl_CreateCommand (interp, "msqlseek",    Msqltcl_Seek,    (ClientData)NULL,
+                     Msqltcl_Kill);
   Tcl_CreateCommand (interp, "msqlmap",     Msqltcl_Map,     (ClientData)NULL,
 		     Msqltcl_Kill);
   Tcl_CreateCommand (interp, "msqlexec",    Msqltcl_Exec,    (ClientData)NULL,
@@ -567,7 +650,7 @@ Msqltcl_Init (interp)
 		     Msqltcl_Kill);
   Tcl_CreateCommand (interp, "msqlresult",  Msqltcl_Result,  (ClientData)NULL,
 		     Msqltcl_Kill);
-  Tcl_CreateCommand (interp, "msqlcol",     Msqltcl_Col, (ClientData)NULL,
+  Tcl_CreateCommand (interp, "msqlcol",     Msqltcl_Col,     (ClientData)NULL,
 		     Msqltcl_Kill);
   Tcl_CreateCommand (interp, "msqlstate",   Msqltcl_State,   (ClientData)NULL,
 		     Msqltcl_Kill);
@@ -575,6 +658,16 @@ Msqltcl_Init (interp)
   /* Initialize msqlstatus global array. */
   clear_msg(interp);
 
+  /* Link the null value element to the corresponding C variable. */
+  if ((MsqlNullvalue = (char*)malloc (12)) == NULL)
+    {
+      fprintf (stderr, "*** msqltcl: out of memory\n") ;
+      return TCL_ERROR ;
+    }
+  (void)strcpy (MsqlNullvalue, MSQL_NULLV_INIT);
+  (void)sprintf (nbuf, "%s(%s)", MsqlStatusArr, MSQL_STATUS_NULLV) ;
+  Tcl_LinkVar (interp, nbuf, (char*)&MsqlNullvalue, TCL_LINK_STRING) ;
+
   /* A little sanity check.
    * If this message appears you must change the source code and recompile.
    */
@@ -612,25 +705,23 @@ Msqltcl_Connect (clientData, interp, argc, argv)
   int        i;
   char       buf[MSQL_BUFF_SIZE];
   int connect ;
-  
-  /* Can't use msql_prologue, handle is 'created' here. */
+
+  /* Pro-forma check (should never fail). */
+  if (msql_prologue (interp, argc, argv, 1, NULL, "?hostname?") < 0)
+    return TCL_ERROR;
 
   /* Find an unused handle. */
-
-    for (i = 0; i < MSQL_HANDLES; i++) {
-	if (MsqlHandle[i].connection <= 0) {
-            hand = i;
-	    break;
-	}
+  for (i = 0; i < MSQL_HANDLES; i++) {
+    if (MsqlHandle[i].connection <= 0) {
+      hand = i;
+      break;
     }
+  }
 
-    if (hand == -1) {
-	Tcl_AppendResult (interp, argv[0], ": no mSQL handles available",
-			  (char*)NULL);
-	return msql_prim_conflict (interp, (char*)NULL) ;
-    }
+  if (hand == -1)
+    return msql_prim_confl ("no mSQL handles available");
 
-  if (argc > 1)
+  if (argc > 1 && strlen (argv[1]) > 0)
     {
       connect = msqlConnect (argv[1]) ;
       strncpy (MsqlHandle[hand].host, argv[1], MSQL_NAME_LEN) ;
@@ -640,34 +731,26 @@ Msqltcl_Connect (clientData, interp, argc, argv)
     {
       if (gethostname (buf, MSQL_NAME_LEN) == 0)
 	{
-	  connect = msqlConnect () ;
+	  connect = msqlConnect ((char*)NULL) ;
 	  strcpy (MsqlHandle[hand].host, buf) ;
 	}
       else
-	{
-	  msql_tcl_perror (interp) ;
-	  Tcl_AppendResult (interp, argv[0], ": cannot get host name",
-			  (char*)NULL);
-	  return TCL_ERROR;
-	}
+	return msql_posix_confl ("gethostname") ;
     }
 
   if (connect < 0)
     {
-      msql_tcl_conflict (interp) ;
       MsqlHandle[hand].connection = -1 ; /* Just to be sure. */
-      Tcl_AppendResult (interp, argv[0], ": connect failed",
-			  (char*)NULL);
-      return TCL_ERROR;
+      return msql_server_confl ();
     }
   else
     MsqlHandle[hand].connection = connect ;
 
-    /* Construct handle and return. */
-    sprintf(buf, "%s%d", MsqlHandlePrefix, hand);
-    Tcl_SetResult(interp, buf, TCL_VOLATILE);
+  /* Construct handle and return. */
+  sprintf(buf, "%s%d", MsqlHandlePrefix, hand);
+  Tcl_SetResult(interp, buf, TCL_VOLATILE);
 
-    return TCL_OK;
+  return TCL_OK;
 }
 
 
@@ -683,32 +766,24 @@ Msqltcl_Connect (clientData, interp, argc, argv)
  */
 
 Msqltcl_Use (clientData, interp, argc, argv)
-    ClientData   clientData;
-    Tcl_Interp  *interp;
-    int          argc;
-    char       **argv;
+     ClientData   clientData;
+     Tcl_Interp  *interp;
+     int          argc;
+     char       **argv;
 {
-    int hand;
-    int toolong ;
+  int hand;
+  
+  if ((hand = msql_prologue(interp, argc, argv, 3, get_handle_conn,
+			    "handle dbname")) < 0)
+    return TCL_ERROR;
 
-    if ((hand = msql_prologue(interp, argc, argv, 3,
-			      " handle dbname", get_handle_conn)) < 0)
-      return TCL_ERROR;
+  if (strlen (argv[2]) >= MSQL_NAME_LEN)
+    return msql_prim_confl ("database name too long") ;
+  if (msqlSelectDB (MsqlHandle[hand].connection, argv[2]) < 0)
+    return msql_server_confl () ;
 
-    toolong = strlen (argv[2]) >= MSQL_NAME_LEN ;
-    if (toolong || msqlSelectDB (MsqlHandle[hand].connection, argv[2]) < 0)
-      {
-	if (toolong)
-	  (void)msql_prim_conflict (interp, "Database name too long") ;
-	else
-	  msql_tcl_conflict (interp) ;
-	Tcl_AppendResult(interp, argv[0], ": msqluse failed", (char*)NULL);
-	return TCL_ERROR;
-      }
-    else
-      strcpy (MsqlHandle[hand].database, argv[2]) ;
-
-    return TCL_OK;
+  strcpy (MsqlHandle[hand].database, argv[2]) ;
+  return TCL_OK;
 }
 
 
@@ -718,7 +793,7 @@ Msqltcl_Use (clientData, interp, argc, argv)
  *
  * Msqltcl_Sel
  *    Implements the msqlsel command:
- *    usage: msqlsql handle sel-query
+ *    usage: msqlsel handle sel-query
  *	                
  *    results:
  *
@@ -727,19 +802,15 @@ Msqltcl_Use (clientData, interp, argc, argv)
  */
 
 Msqltcl_Sel (clientData, interp, argc, argv)
-    ClientData   clientData;
-    Tcl_Interp  *interp;
-    int          argc;
-    char       **argv;
+     ClientData   clientData;
+     Tcl_Interp  *interp;
+     int          argc;
+     char       **argv;
 {
   int     hand;
-  char    buf[MSQL_BUFF_SIZE];
-  char    conv_buf[20];
-  int     text_buf_size;
-
-
-  if ((hand = msql_prologue(interp,argc, argv, 3,
-			    " handle sel-query", get_handle_db)) < 0)
+  
+  if ((hand = msql_prologue(interp, argc, argv, 3, get_handle_db,
+			    "handle sel-query")) < 0)
     return TCL_ERROR;
 
   /* Flush any previous result. */
@@ -750,22 +821,17 @@ Msqltcl_Sel (clientData, interp, argc, argv)
     }
 
   if (msqlQuery (MsqlHandle[hand].connection, argv[2]) < 0)
-    {
-      msql_tcl_conflict (interp) ;
-      Tcl_SetResult(interp, "msqlsel failed", TCL_VOLATILE);
-      return TCL_ERROR;
-    }
+    return msql_server_confl () ;
 
   if ((MsqlHandle[hand].result = msqlStoreResult ()) == NULL)
     {
-      msql_tcl_conflict (interp) ;
-      Tcl_SetResult(interp, "could not get result", TCL_VOLATILE);
-      return TCL_ERROR;
+      (void)strcpy (interp->result, "-1") ;
     }
   else
     {
       MsqlHandle[hand].res_count = msqlNumRows (MsqlHandle[hand].result) ;
       MsqlHandle[hand].col_count = msqlNumFields (MsqlHandle[hand].result) ;
+      (void)sprintf (interp->result, "%d", MsqlHandle[hand].res_count) ;
     }
 
   return TCL_OK;
@@ -793,25 +859,21 @@ Msqltcl_Exec (clientData, interp, argc, argv)
 {
   int     hand;
   
-  if ((hand = msql_prologue(interp, argc, argv, 3,
-			    " handle sql-statement", get_handle_db)) < 0)
-      return TCL_ERROR;
+  if ((hand = msql_prologue(interp, argc, argv, 3, get_handle_db,
+			    "handle sql-statement")) < 0)
+    return TCL_ERROR;
 
-    /* Flush any previous result. */
-    if (MsqlHandle[hand].result != NULL)
-      {
-	msqlFreeResult (MsqlHandle[hand].result) ;
-	MsqlHandle[hand].result = NULL ;
-      }
+  /* Flush any previous result. */
+  if (MsqlHandle[hand].result != NULL)
+    {
+      msqlFreeResult (MsqlHandle[hand].result) ;
+      MsqlHandle[hand].result = NULL ;
+    }
 
-    if (msqlQuery (MsqlHandle[hand].connection, argv[2]) < 0)
-      {
-	msql_tcl_conflict (interp) ;
-	Tcl_SetResult(interp, "msqlexec failed", TCL_VOLATILE);
-	return TCL_ERROR;
-      }
+  if (msqlQuery (MsqlHandle[hand].connection, argv[2]) < 0)
+    return msql_server_confl () ;
 
-    return TCL_OK ;
+  return TCL_OK ;
 }
 
 
@@ -827,34 +889,95 @@ Msqltcl_Exec (clientData, interp, argc, argv)
  */
 
 Msqltcl_Next (clientData, interp, argc, argv)
+     ClientData   clientData;
+     Tcl_Interp  *interp;
+     int          argc;
+     char       **argv;
+{
+  int hand;
+  int idx ;
+  m_row row ;
+  char* val ;
+  
+  if ((hand = msql_prologue(interp, argc, argv, 2, get_handle_res,
+			    "handle")) < 0)
+    return TCL_ERROR;
+
+  
+  if (MsqlHandle[hand].res_count == 0)
+    return TCL_OK ;
+  else if ((row = msqlFetchRow (MsqlHandle[hand].result)) == NULL)
+    {
+      MsqlHandle[hand].res_count = 0 ;
+      return msql_prim_confl ("result counter out of sync") ;
+    }
+  else
+    MsqlHandle[hand].res_count-- ;
+  
+  for (idx = 0 ; idx < MsqlHandle[hand].col_count ; idx++)
+    {
+      if ((val = *row++) == NULL)
+	val = MsqlNullvalue ;
+      Tcl_AppendElement (interp, val) ;
+    }
+  
+  return TCL_OK;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Msqltcl_Seek
+ *    Implements the msqlseek command:
+ *    usage: msqlseek handle rownumber
+ *	                
+ *    results:
+ *	number of remaining rows
+ */
+
+Msqltcl_Seek (clientData, interp, argc, argv)
     ClientData   clientData;
     Tcl_Interp  *interp;
     int          argc;
     char       **argv;
 {
     int hand;
-    int idx ;
-    char buf[MSQL_BUFF_SIZE];
-    m_row row ;
-
-    if ((hand = msql_prologue(interp, argc, argv, 2,
-			      " handle", get_handle_res)) < 0)
+    int res;
+    int row;
+    int total;
+   
+    if ((hand = msql_prologue(interp, argc, argv, 3, get_handle_res,
+                              " handle row-index")) < 0)
       return TCL_ERROR;
 
-    if (MsqlHandle[hand].res_count == 0)
-      return TCL_OK ;
-    else if ((row = msqlFetchRow (MsqlHandle[hand].result)) == NULL)
+    if ((res = Tcl_GetInt (interp, argv[2], &row)) != TCL_OK)
+      return res;
+    
+    total = msqlNumRows (MsqlHandle[hand].result);
+    
+    if (total + row < 0)
       {
-	MsqlHandle[hand].res_count = 0 ;
-	Tcl_AppendResult(interp, argv[0], ": result counter out of sync", (char*)NULL);
-	return msql_prim_conflict (interp, (char*)NULL) ;
+	msqlDataSeek (MsqlHandle[hand].result, 0);
+	MsqlHandle[hand].res_count = total;
+      }
+    else if (row < 0)
+      {
+	msqlDataSeek (MsqlHandle[hand].result, total + row);
+	MsqlHandle[hand].res_count = -row;
+      }
+    else if (row >= total)
+      {
+	msqlDataSeek (MsqlHandle[hand].result, row);
+	MsqlHandle[hand].res_count = 0;
       }
     else
-      MsqlHandle[hand].res_count-- ;
+      {
+	msqlDataSeek (MsqlHandle[hand].result, row);
+	MsqlHandle[hand].res_count = total - row;
+      }
 
-    for (idx = 0 ; idx < MsqlHandle[hand].col_count ; idx++)
-      Tcl_AppendElement (interp, *row++) ;
-    
+    (void)sprintf (interp->result, "%d", MsqlHandle[hand].res_count) ;
     return TCL_OK;
 }
 
@@ -888,9 +1011,10 @@ Msqltcl_Map (clientData, interp, argc, argv)
   int listArgc ;
   char** listArgv ;
   m_row row ;
-
-  if ((hand = msql_prologue(interp, argc, argv, 4,
-			    " handle binding-list script", get_handle_res)) < 0)
+  char* val ;
+  
+  if ((hand = msql_prologue(interp, argc, argv, 4, get_handle_res,
+			    "handle binding-list script")) < 0)
     return TCL_ERROR;
 
   if (Tcl_SplitList (interp, argv[2], &listArgc, &listArgv) != TCL_OK)
@@ -898,10 +1022,8 @@ Msqltcl_Map (clientData, interp, argc, argv)
   
   if (listArgc > MsqlHandle[hand].col_count)
     {
-      Tcl_AppendResult (interp, argv[0], ": too many variables in binding list",
-			(char*)NULL) ;
       ckfree ((char*)listArgv) ;
-      return msql_prim_conflict (interp, (char*)NULL) ;
+      return msql_prim_confl ("too many variables in binding list") ;
     }
   else
     count = (listArgc < MsqlHandle[hand].col_count)?listArgc
@@ -913,10 +1035,8 @@ Msqltcl_Map (clientData, interp, argc, argv)
       if ((row = msqlFetchRow (MsqlHandle[hand].result)) == NULL)
 	{
 	  MsqlHandle[hand].res_count = 0 ;
-	  Tcl_AppendResult(interp, argv[0], ": result counter out of sync",
-			   (char*)NULL);
 	  ckfree ((char*)listArgv) ;
-	  return msql_prim_conflict (interp, (char*)NULL) ;
+	  return msql_prim_confl ("result counter out of sync") ;
 	}
       else
 	MsqlHandle[hand].res_count-- ;
@@ -926,7 +1046,9 @@ Msqltcl_Map (clientData, interp, argc, argv)
 	{
 	  if (listArgv[idx][0] != '-')
 	    {
-	      if (Tcl_SetVar (interp, listArgv[idx], *row++, TCL_LEAVE_ERR_MSG)
+	      if ((val = *row++) == NULL)
+		val = MsqlNullvalue ;
+	      if (Tcl_SetVar (interp, listArgv[idx], val, TCL_LEAVE_ERR_MSG)
 		  == NULL)
 		{
 		  ckfree ((char*)listArgv) ;
@@ -962,8 +1084,8 @@ Msqltcl_Map (clientData, interp, argc, argv)
  *----------------------------------------------------------------------
  *
  * Msqltcl_Info
- * Implements the msqldb command:
- * usage: msqldb handle option
+ * Implements the msqlinfo command:
+ * usage: msqlinfo handle option
  *
  */
 
@@ -974,16 +1096,17 @@ Msqltcl_Info (clientData, interp, argc, argv)
      int          argc;
      char       **argv;
 {
+  char buf[MSQL_BUFF_SIZE];
   int count ;
   int hand ;
   int idx ;
   m_result* list ;
   m_row row ;
-  
+  char* val ;
   
   /* We can't fully check the handle at this stage. */
-  if ((hand = msql_prologue(interp, argc, argv, 3,
-			    " handle option", get_handle_plain)) < 0)
+  if ((hand = msql_prologue(interp, argc, argv, 3, get_handle_plain,
+			    "handle option")) < 0)
     return TCL_ERROR;
 
   for (idx = 0;
@@ -995,10 +1118,10 @@ Msqltcl_Info (clientData, interp, argc, argv)
     {
     case MSQL_INFNAME_OPT:
     case MSQL_INFTABLES_OPT:
-      hand = get_handle_db (interp, argv[1]) ;
+      hand = get_handle_db (argv[1]) ;
       break ;
     case MSQL_INFNAMEQ_OPT:
-      if ((hand = get_handle_conn (interp, argv[1])) >= 0)
+      if ((hand = get_handle_conn (argv[1])) >= 0)
 	{
 	  if (MsqlHandle[hand].database[0] == '\0')
 	    return TCL_OK ; /* Return empty string if no current db. */
@@ -1006,25 +1129,19 @@ Msqltcl_Info (clientData, interp, argc, argv)
       break ;
     case MSQL_INFHOST_OPT:
     case MSQL_INFLIST_OPT:
-      hand = get_handle_conn (interp, argv[1]) ;
+      hand = get_handle_conn (argv[1]) ;
       break ;
     case MSQL_INFHOSTQ_OPT:
       if (MsqlHandle[hand].connection < 0)
 	return TCL_OK ; /* Return empty string if not connected. */
       break ;
     default: /* unknown option */
-      Tcl_AppendResult (interp, argv[0], ": '", argv[2], "' unknown option",
-			(char*)NULL);
-      return msql_prim_conflict (interp, (char*)NULL) ;
+      sprintf (buf, "'%s' unknown option", argv[2]);
+      return msql_prim_confl (buf) ;
     }
 
   if (hand < 0)
-    {
-      /* Status has already been set by the handle checking function. */
-      Tcl_AppendResult (interp, argv[0], ": ", argv[1], " invalid handle",
-			(char*)NULL) ;
       return TCL_ERROR ;
-    }
 
   /* Handle OK, return the requested info. */
   switch (idx)
@@ -1035,15 +1152,12 @@ Msqltcl_Info (clientData, interp, argc, argv)
       break ;
     case MSQL_INFTABLES_OPT:
       if ((list = msqlListTables (MsqlHandle[hand].connection)) == NULL)
-	{
-	  Tcl_AppendResult (interp, argv[0], ": could not access table names",
-			    "; server may have gone away", (char*)NULL);
-	  return msql_prim_conflict (interp, (char*)NULL) ;
-	}
+	return msql_prim_confl ("could not access table names; server may have gone away") ;
+
       for (count = msqlNumRows (list); count > 0; count--)
 	{
-	  row = msqlFetchRow (list) ;
-	  Tcl_AppendElement (interp, *row) ;
+	  val = *(row = msqlFetchRow (list)) ;
+	  Tcl_AppendElement (interp, (val == NULL)?"":val) ;
 	}
       msqlFreeResult (list) ;
       break ;
@@ -1053,22 +1167,17 @@ Msqltcl_Info (clientData, interp, argc, argv)
       break ;
     case MSQL_INFLIST_OPT:
       if ((list = msqlListDBs (MsqlHandle[hand].connection)) == NULL)
-	{
-	  Tcl_AppendResult (interp, argv[0], ": could not access database names",
-			    "; server may have gone away", (char*)NULL) ;
-	  return msql_prim_conflict (interp, (char*)NULL) ;
-	}
+	return msql_prim_confl ("could not access database names; server may have gone away") ;
+
       for (count = msqlNumRows (list); count > 0; count--)
 	{
-	  row = msqlFetchRow (list) ;
-	  Tcl_AppendElement (interp, *row) ;
+	  val = *(row = msqlFetchRow (list)) ;
+	  Tcl_AppendElement (interp, (val == NULL)?"":val) ;
 	}
       msqlFreeResult (list) ;
       break ;
     default: /* should never happen */
-      Tcl_AppendResult (interp, argv[0], ": weirdness in Msqltcl_Info",
-			(char*)NULL);
-      return msql_prim_conflict (interp, (char*)NULL) ;
+      return msql_prim_confl ("weirdness in Msqltcl_Info") ;
     }
   return TCL_OK ;
 }
@@ -1090,6 +1199,7 @@ Msqltcl_Result (clientData, interp, argc, argv)
      int          argc;
      char       **argv;
 {
+  char buf[MSQL_BUFF_SIZE];
   int count ;
   int hand ;
   int idx ;
@@ -1098,8 +1208,8 @@ Msqltcl_Result (clientData, interp, argc, argv)
 
 
   /* We can't fully check the handle at this stage. */
-  if ((hand = msql_prologue(interp, argc, argv, 3,
-			    " handle option", get_handle_plain)) < 0)
+  if ((hand = msql_prologue(interp, argc, argv, 3, get_handle_plain,
+			    " handle option")) < 0)
     return TCL_ERROR;
 
   for (idx = 0;
@@ -1111,29 +1221,25 @@ Msqltcl_Result (clientData, interp, argc, argv)
     {
     case MSQL_RESROWS_OPT:
     case MSQL_RESCOLS_OPT:
-      hand = get_handle_res (interp, argv[1]) ;
+    case MSQL_RESCUR_OPT:
+      hand = get_handle_res (argv[1]) ;
       break ;
     case MSQL_RESROWSQ_OPT:
     case MSQL_RESCOLSQ_OPT:
-      if ((hand = get_handle_db (interp, argv[1])) >= 0)
+    case MSQL_RESCURQ_OPT:
+      if ((hand = get_handle_db (argv[1])) >= 0)
 	{
 	  if (MsqlHandle[hand].result == NULL)
-	    return TCL_OK ; /* Return null string if no pending result. */
+	    return TCL_OK ; /* Return empty string if no pending result. */
 	}
       break ;
     default: /* unknown option */
-      Tcl_AppendResult (interp, argv[0], ": '", argv[2], "' unknown option",
-			(char*)NULL);
-      return msql_prim_conflict (interp, (char*)NULL) ;
+      sprintf (buf, "'%s' unknown option", argv[2]);
+      return msql_prim_confl (buf) ;
     }
 
   if (hand < 0)
-    {
-      /* Status has already been set by the handle checking function. */
-      Tcl_AppendResult (interp, argv[0], ": ", argv[1], " invalid handle",
-			(char*)NULL) ;
-      return TCL_ERROR ;
-    }
+    return TCL_ERROR ;
 
   /* Handle OK; return requested info. */
   switch (idx)
@@ -1146,6 +1252,10 @@ Msqltcl_Result (clientData, interp, argc, argv)
     case MSQL_RESCOLSQ_OPT:
       sprintf (interp->result, "%d", MsqlHandle[hand].col_count) ;
       break ;
+    case MSQL_RESCUR_OPT:
+    case MSQL_RESCURQ_OPT:
+      sprintf (interp->result, "%d", msqlNumRows (MsqlHandle[hand].result)
+	       - MsqlHandle[hand].res_count) ;
     default:
     }
   return TCL_OK ;
@@ -1174,6 +1284,7 @@ Msqltcl_Col (clientData, interp, argc, argv)
      int          argc;
      char       **argv;
 {
+  char buf[MSQL_BUFF_SIZE];
   int coln ;
   int conflict ;
   int current_db ;
@@ -1186,9 +1297,9 @@ Msqltcl_Col (clientData, interp, argc, argv)
   char* sep ;
   int simple ;
   
-  if ((hand = msql_prologue(interp, argc, argv, 4,
-			    " handle table-name option ?option ...?",
-			    get_handle_db)) < 0)
+  /* This check is enough only without '-current'. */
+  if ((hand = msql_prologue(interp, argc, argv, 4, get_handle_db,
+			    "handle table-name option ?option ...?")) < 0)
     return TCL_ERROR;
 
   /* Fetch column info.
@@ -1198,12 +1309,8 @@ Msqltcl_Col (clientData, interp, argc, argv)
   
   if (current_db)
     {
-      if ((hand = get_handle_res (interp, argv[1])) < 0)
-	{
-	  Tcl_AppendResult (interp, argv[0], " ", argv[1],
-			    ": invalid handle ", (char*)NULL);
-	  return TCL_ERROR ;
-	}
+      if ((hand = get_handle_res (argv[1])) < 0)
+	return TCL_ERROR ;
       else
 	result = MsqlHandle[hand].result ;
     }
@@ -1211,10 +1318,9 @@ Msqltcl_Col (clientData, interp, argc, argv)
     {
       if ((result = msqlListFields (MsqlHandle[hand].connection, argv[2])) == NULL)
 	{
-	  Tcl_AppendResult (interp, argv[0], ": no column info for table '",
-			    argv[2], "'",
-			    "; server may have gone away", (char*)NULL) ;
-	  return msql_prim_conflict (interp, (char*)NULL) ;
+	  sprintf (buf, "no column info for table '%s'; %s", argv[2],
+		   "server may have gone away") ;
+	  return msql_prim_confl (buf) ;
 	}
     }
 
@@ -1255,7 +1361,7 @@ Msqltcl_Col (clientData, interp, argc, argv)
     }
   else
     {
-      ckfree ((char*)listArgv) ; /* listArgv == 1, no splitting */
+      ckfree ((char*)listArgv) ; /* listArgc == 1, no splitting */
       for (idx = 3, sep = "{"; idx < argc; idx++, sep = " {")
 	{
 	  Tcl_AppendResult (interp, sep, (char*)NULL) ;
@@ -1274,7 +1380,7 @@ Msqltcl_Col (clientData, interp, argc, argv)
   
   if (!current_db)
     msqlFreeResult (result) ;
-  return TCL_OK ;
+  return (conflict)?TCL_ERROR:TCL_OK ;
 }
 
 
@@ -1298,17 +1404,12 @@ Msqltcl_State (clientData, interp, argc, argv)
   int numeric ;
   char* res ;
   
-  if (argc < 2)
-    {
-      Tcl_AppendResult (interp, argv[0], ": too few arguments", (char*)NULL);
-      return msql_prim_conflict (interp, (char*)NULL) ;
-    }
-  else if ((numeric = (strcmp (argv[1], "-numeric") == 0)) && argc < 3)
-    {
-      Tcl_AppendResult (interp, argv[0], ": too few arguments", (char*)NULL);
-      return msql_prim_conflict (interp, (char*)NULL) ;
-    }
+  if (msql_prologue(interp, argc, argv, 2, NULL, "?-numeric? handle") < 0)
+    return TCL_ERROR;
 
+  if ((numeric = (strcmp (argv[1], "-numeric") == 0)) && argc < 3)
+    return msql_prim_confl ("handle required") ;
+  
   hp = (numeric)?argv[2]:argv[1] ;
 
   if (HSYNTAX(hp,hi) < 0)
@@ -1322,7 +1423,7 @@ Msqltcl_State (clientData, interp, argc, argv)
   else
     res = (numeric)?"4":"RESULT_PENDING" ;
 
-  strcpy (interp->result, res) ;
+  (void)strcpy (interp->result, res) ;
   return TCL_OK ;
 }
 
@@ -1346,14 +1447,15 @@ Msqltcl_Close (clientData, interp, argc, argv)
 {
   int     hand;
   
+  /* If handle omitted, close all connections. */
   if (argc == 1)
     {
       Msqltcl_Kill ((ClientData)NULL) ;
       return TCL_OK ;
     }
   
-  if ((hand = msql_prologue(interp, argc, argv, 2,
-			    " handle", get_handle_conn)) < 0)
+  if ((hand = msql_prologue(interp, argc, argv, 2, get_handle_conn,
+			    "handle")) < 0)
     return TCL_ERROR;
 
   msqlClose (MsqlHandle[hand].connection) ;
